@@ -12,9 +12,7 @@ class Learner:
     """
     Learner (based on Soft Actor-Critic)
     =================
-
     Paper: https://arxiv.org/pdf/1812.05905.pdf
-
     Attributes:
         env: the instance of environment object
         max_steps (int): maximum number of interactions do in environment
@@ -33,7 +31,6 @@ class Learner:
         save_path (str): path to the models for saving
         db_path (str): path to the database
         logging_wandb (bool): logging by WanDB
-        log_freq (int): logging frequency
     """
 
     def __init__(
@@ -61,16 +58,21 @@ class Learner:
         db_path: str = None,
         # ---
         logging_wandb: bool = False,
-        log_freq: int = 64,
     ):
         self._env = env
         self._max_steps = max_steps
+        self._gradient_steps = gradient_steps
         self._learning_starts = learning_starts
         self._gamma = tf.constant(gamma)
         self._tau = tf.constant(tau)
-        self._save_path = save_path
         self._logging_wandb = logging_wandb
-        self._log_freq = log_freq
+        self._save_path = save_path
+
+        # logging metrics
+        self._loss_a = tf.keras.metrics.Mean()
+        self._loss_c1 = tf.keras.metrics.Mean()
+        self._loss_c2 = tf.keras.metrics.Mean()
+        self._loss_alpha = tf.keras.metrics.Mean()
 
         # init param 'alpha' - Lagrangian constraint
         self._log_alpha = tf.Variable(0.0, trainable=True, name="log_alpha")
@@ -127,8 +129,18 @@ class Learner:
         else:
             checkpointer = reverb.checkpointers.DefaultCheckpointer(path=db_path)
 
+        # actual training step
+        self._train_step = tf.Variable(
+            0,
+            trainable=False,
+            dtype=tf.int32,
+            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
+            shape=(),
+        )
+
         # prepare variable container
         self._variables_container = {
+            "train_step": self._train_step,
             "policy_variables": self._actor.model.variables,
         }
 
@@ -138,9 +150,6 @@ class Learner:
             self._variables_container,
         )
 
-        # Ratio for samples per insert rate limiting tolerance
-        SAMPLES_PER_INSERT_TOLERANCE_RATIO = 0.1
-
         # Initialize the reverb server
         self.server = reverb.Server(
             tables=[
@@ -148,15 +157,7 @@ class Learner:
                     name="experience",
                     sampler=reverb.selectors.Uniform(),
                     remover=reverb.selectors.Fifo(),
-                    rate_limiter=reverb.rate_limiters.SampleToInsertRatio(
-                        min_size_to_sample=learning_starts,
-                        samples_per_insert=gradient_steps,
-                        error_buffer=(
-                            learning_starts
-                            * SAMPLES_PER_INSERT_TOLERANCE_RATIO
-                            * gradient_steps
-                        ),
-                    ),
+                    rate_limiter=reverb.rate_limiters.MinSize(learning_starts),
                     max_size=buffer_capacity,
                     max_times_sampled=0,
                     signature={
@@ -193,13 +194,11 @@ class Learner:
         # Initializes the reverb client
         self.client = reverb.Client("localhost:8000")
         self.tf_client = reverb.TFClient(server_address="localhost:8000")
-        self._dataset_iterator = iter(
-            reverb.TrajectoryDataset.from_table_signature(
-                server_address="localhost:8000",
-                table="experience",
-                max_in_flight_samples_per_worker=10,
-            ).batch(batch_size)
-        )
+        self._dataset = reverb.TrajectoryDataset.from_table_signature(
+            server_address="localhost:8000",
+            table="experience",
+            max_in_flight_samples_per_worker=10,
+        ).batch(batch_size)
 
         # init Weights & Biases
         if self._logging_wandb:
@@ -235,18 +234,18 @@ class Learner:
 
     # -------------------------------- update critic ------------------------------- #
     def _update_critic(self, batch):
-        next_action, next_log_pi = self._actor.predict(batch["next_observation"])
+        next_action, next_log_pi = self._actor.predict(batch.data["next_observation"])
 
         # target Q-values
-        next_q_1 = self._critic_targ_1.model([batch["next_observation"], next_action])
-        next_q_2 = self._critic_targ_2.model([batch["next_observation"], next_action])
+        next_q_1 = self._critic_targ_1.model([batch.data["next_observation"], next_action])
+        next_q_2 = self._critic_targ_2.model([batch.data["next_observation"], next_action])
         next_q = tf.minimum(next_q_1, next_q_2)
         # tf.print(f'nextQ: {next_q.shape}')
 
         # Bellman Equation
         Q_targets = tf.stop_gradient(
-            batch["reward"]
-            + (1.0 - batch["terminal"])
+            batch.data["reward"]
+            + (1 - batch.data["terminal"])
             * self._gamma
             * (next_q - self._alpha * next_log_pi)
         )
@@ -254,7 +253,7 @@ class Learner:
 
         # update critic '1'
         with tf.GradientTape() as tape:
-            q_values = self._critic_1.model([batch["observation"], batch["action"]])
+            q_values = self._critic_1.model([batch.data["observation"], batch.data["action"]])
             q_losses = tf.losses.huber(  # less sensitive to outliers in batch
                 y_true=Q_targets, y_pred=q_values
             )
@@ -268,7 +267,7 @@ class Learner:
 
         # update critic '2'
         with tf.GradientTape() as tape:
-            q_values = self._critic_2.model([batch["observation"], batch["action"]])
+            q_values = self._critic_2.model([batch.data["observation"], batch.data["action"]])
             q_losses = tf.losses.huber(  # less sensitive to outliers in batch
                 y_true=Q_targets, y_pred=q_values
             )
@@ -285,12 +284,12 @@ class Learner:
     def _update_actor(self, batch):
         with tf.GradientTape() as tape:
             # predict action
-            y_pred, log_pi = self._actor.predict(batch["observation"])
+            y_pred, log_pi = self._actor.predict(batch.data["observation"])
             # tf.print(f'log_pi: {log_pi.shape}')
 
             # predict q value
-            q_1 = self._critic_1.model([batch["observation"], y_pred])
-            q_2 = self._critic_2.model([batch["observation"], y_pred])
+            q_1 = self._critic_1.model([batch.data["observation"], y_pred])
+            q_2 = self._critic_2.model([batch.data["observation"], y_pred])
             q = tf.minimum(q_1, q_2)
             # tf.print(f'q: {q.shape}')
 
@@ -307,7 +306,7 @@ class Learner:
 
     # -------------------------------- update alpha ------------------------------- #
     def _update_alpha(self, batch):
-        _, log_pi = self._actor.predict(batch["observation"])
+        _, log_pi = self._actor.predict(batch.data["observation"])
         # tf.print(f'y_pred: {y_pred.shape}')
         # tf.print(f'log_pi: {log_pi.shape}')
 
@@ -326,63 +325,69 @@ class Learner:
 
     @tf.function
     def _update(self):
-        # Get data from replay buffer
-        sample = next(self._dataset_iterator).data
+        #  env_steps : gradient_steps = 1 : 1
+        for sample in self._dataset.take(self._gradient_steps):
+            # re-new noise matrix every update of 'log_std' params
+            self._actor.reset_noise()
 
-        # re-new noise matrix every update of 'log_std' params
-        self._actor.reset_noise()
+            # Alpha param update
+            self._loss_alpha.update_state(self._update_alpha(sample))
 
-        # Alpha param update
-        loss_alpha = self._update_alpha(sample)
+            l_c1, l_c2 = self._update_critic(sample)
+            self._loss_c1.update_state(l_c1)
+            self._loss_c2.update_state(l_c2)
 
-        # Critic model update
-        loss_c1, loss_c2 = self._update_critic(sample)
+            # Actor model update
+            self._loss_a.update_state(self._update_actor(sample))
 
-        # Actor model update
-        loss_a = self._update_actor(sample)
-
-        # -------------------- soft update target networks -------------------- #
-        self._update_target(self._critic_1, self._critic_targ_1, tau=self._tau)
-        self._update_target(self._critic_2, self._critic_targ_2, tau=self._tau)
+            # -------------------- soft update target networks -------------------- #
+            self._update_target(self._critic_1, self._critic_targ_1, tau=self._tau)
+            self._update_target(self._critic_2, self._critic_targ_2, tau=self._tau)
 
         # store new actor's params
         self._push_variables()
 
-        return loss_alpha, loss_c1, loss_c2, loss_a
-
     def run(self):
-        for step in range(self._learning_starts, self._max_steps, 1):
+        for step in range(self._learning_starts, self._max_steps, self._gradient_steps):
+            # update train_step (otlacok modelov)
+            self._train_step.assign(step)
+
             # update models
-            loss_alpha, loss_c1, loss_c2, loss_a = self._update()
+            self._update()
 
             # log metrics
-            self._logging_models(loss_alpha, loss_c1, loss_c2, loss_a, step)
+            self._logging_models()
 
-    def _logging_models(self, loss_alpha, loss_c1, loss_c2, loss_a, step):
-        # log into console
-        if step % self._log_freq == 0:
-            print("=============================================")
-            print(f"Step: {step}")
-            print(f"Actor's loss: {loss_a}")
-            print(f"Critic 1's loss: {loss_c1}")
-            print(f"Critic 2's loss: {loss_c2}")
-            print("=============================================")
-            print(
-                f"Training ... {math.floor(step * 100.0 / self._max_steps)} %"
-            )  # noqa
-
-        # log into wandb cloud
+    def _logging_models(self):
+        print("=============================================")
+        print(f"Step: {self._train_step.numpy()}")
+        print(f"Actor's loss: {self._loss_a.result()}")
+        print(f"Critic 1's loss: {self._loss_c1.result()}")
+        print(f"Critic 2's loss: {self._loss_c2.result()}")
+        print("=============================================")
+        print(
+            f"Training ... {math.floor(self._train_step.numpy() * 100.0 / self._max_steps)} %"  # noqa
+        )
         if self._logging_wandb:
+            # logging of epoch's mean loss
             wandb.log(
                 {
-                    "loss_a": loss_a,
-                    "loss_c1": loss_c1,
-                    "loss_c2": loss_c2,
-                    "loss_alpha": loss_alpha,
+                    "loss_a": self._loss_a.result(),
+                    "loss_c1": self._loss_c1.result(),
+                    "loss_c2": self._loss_c2.result(),
+                    "loss_alpha": self._loss_alpha.result(),
                     "alpha": self._alpha,
                 },
-                step=step,
+                step=self._train_step.numpy(),
             )
+        self._clear_metrics()  # clear stored metrics of losses
+
+    def _clear_metrics(self):
+        # reset logger
+        self._loss_a.reset_states()
+        self._loss_c1.reset_states()
+        self._loss_c2.reset_states()
+        self._loss_alpha.reset_states()
 
     def save(self):
         if self._save_path is not None:
@@ -402,4 +407,4 @@ class Learner:
 
         # Save the model.
         with open("model_A.tflite", "wb") as f:
-            f.write(tflite_model)
+            f.write(tflite_model)
\ No newline at end of file
